/*****************************************************************************
 *
 * libwsw, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ejabberd@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"

#include <gloox.h>
#pragma warning(push)
#pragma warning(disable: 4800)
#include <client.h>
#pragma warning(pop)
#include <stanza.h>
#include <iqhandler.h>

#include "gloox_drv.h"
#include "../../utils.h"
//------------------------------------------------------------------------------
using namespace gloox;
using namespace WSW::Core;
using namespace WSW::DateTime;
using namespace WSW::Drivers::XMPP;
//------------------------------------------------------------------------------
namespace
  {
//--//----------------------------------------------------------------------------
  CollectionLink parseCollectionLink(Tag* tag)
    {
    CollectionLink link;

    if (tag)
      {
      if (tag->hasAttribute("start"))
        link.start = str2DateTime(tag->findAttribute("start"));
      link.with = tag->findAttribute("with");
      }

    return link;
    }
//--//----------------------------------------------------------------------------
  CollectionInfo parseCollectionInfo(Tag* chat)
    {
    CollectionInfo info;

    if (chat)
      {
      if (chat->hasAttribute("start"))
        info.self.start = str2DateTime(chat->findAttribute("start"));

      info.self.with = chat->findAttribute("with");
      info.subject = chat->findAttribute("subject");
      info.thread = chat->findAttribute("thread");
      std::string crypt = chat->findAttribute("crypt");
      info.crypt = (crypt == "true" || crypt == "1");

      info.prev = parseCollectionLink(chat->findChild("previous"));
      info.next = parseCollectionLink(chat->findChild("next"));
      }

    return info;
    }
//--//----------------------------------------------------------------------------
  Message parseMessageInfo(Tag* msg_tag, const TimeType& start)
    {
    Message msg;

    if (msg_tag)
      {
      if (msg_tag->name() == "from" || msg_tag->name() == "to")
        msg.body = msg_tag->findChild("body")->cdata();
      else
        msg.body = msg_tag->cdata();
      
      if (msg_tag->hasAttribute("utc"))
        msg.utc = str2DateTime(msg_tag->findAttribute("utc"));
      else if (msg_tag->hasAttribute("secs"))
        msg.utc = start + boost::posix_time::seconds(boost::lexical_cast<unsigned long>(msg_tag->findAttribute("secs")));
      
      if (msg_tag->name() == "from")
        msg.direction = Message::FROM;
      else if (msg_tag->name() == "to")
        msg.direction = Message::TO;
      else
        {
        assert(msg_tag->name() == "note");
        msg.direction = Message::NOTE;
        }

      if (msg_tag->hasAttribute("name"))
        msg.name = msg_tag->findAttribute("name");
      }

    return msg;
    }
//--//----------------------------------------------------------------------------
  RSM_Result parseRSM(Tag* rsm)
    {
    RSM_Result rsm_res;

    if (rsm)
      {
      if (rsm->hasChild("first"))
        {
        Tag* first = rsm->findChild("first");
        rsm_res.first = first->cdata();
        if (first->hasAttribute("index"))
          rsm_res.first_index = boost::lexical_cast<unsigned long>(first->findAttribute("index"));
        }

      if (rsm->hasChild("last"))
        rsm_res.last = rsm->findChild("last")->cdata();

      if (rsm->hasChild("count"))
        rsm_res.count = boost::lexical_cast<unsigned long>(rsm->findChild("count")->cdata());
      }

    return rsm_res;
    }
//--//----------------------------------------------------------------------------
  void addRSM(Tag* cmd, const RSM_Query& rsm)
    {
    // Do not add RSM node at all if nothing to specify there
    if (rsm.mode == RSM_Query::NONE && rsm.limit == RSM_Query::NO_LIMIT)
      return;

    Tag* rsm_tag = new Tag("set");
    rsm_tag->addAttribute("xmlns", XMLNS_RESULT_SET_MANAGEMENT);

    if (rsm.mode != RSM_Query::NONE)
      {
      Tag* node = new Tag(rsm.getModeAsString());
      node->addCData(rsm.value);
      rsm_tag->addChild(node);
      }

    if (rsm.limit != RSM_Query::NO_LIMIT)
      {
      Tag* node = new Tag("max");
      node->addCData(boost::lexical_cast<std::string>(rsm.limit));
      rsm_tag->addChild(node);
      }

    cmd->addChild(rsm_tag);
    }
//--//----------------------------------------------------------------------------
  Tag* createLinkTag(const char* tag_name, const CollectionLink& link)
    {
    Tag* tag = new Tag(tag_name);
    
    tag->addAttribute("with", link.with);
    tag->addAttribute("start", dateTime2Str(link.start));

    return tag;
    }
//--//----------------------------------------------------------------------------
  void sendIqCmd(Client* client, Tag* cmd, IqHandler* handler, StanzaSubType iq_type = StanzaIqGet)
    {
    std::string id = client->getID();
    Stanza *req = Stanza::createIqStanza(client->jid().bare(), id, iq_type);
    req->addChild(cmd);
    client->trackID(handler, id, 0);
    client->send(req);
    }
//--//----------------------------------------------------------------------------
  }
//------------------------------------------------------------------------------
class IqHandlerRetrieveCollections: public IqHandler
  {
  public:
    IqHandlerRetrieveCollections(const Driver::ReceiveCollections& func): func_(func) {}

    ~IqHandlerRetrieveCollections() {}

    // Shouldn't ever be called.
    virtual bool handleIq(Stanza *stanza) { return false; }

    virtual bool handleIqID(Stanza* stanza, int context)
      {
      if (stanza->error() == StanzaErrorUndefined)
        {
        Tag* list_cmd = stanza->findChild("list");
        Tag::TagList children = list_cmd->children();
        Driver::Collections infos;

        for (Tag::TagList::const_iterator it = children.begin(); it != children.end(); ++it)
          if ((*it)->name() == "chat")
            infos.push_back(parseCollectionInfo(*it));

        func_(infos, parseRSM(list_cmd->findChild("set", "xmlns", XMLNS_RESULT_SET_MANAGEMENT)), OK);
        }
      else
        func_(Driver::Collections(), RSM_Result(), SERVER_ERROR);

      delete this; // Not that nice looking, but simple & effective ;-)
      return true;
      }

  private:
    Driver::ReceiveCollections func_;
  };
//------------------------------------------------------------------------------
class IqHandlerRetrieveMessages: public IqHandler
  {
  public:
    IqHandlerRetrieveMessages(const Driver::ReceiveMessages& func_msgs, const Driver::ReceiveCollection& func_coll):
      func_msgs_(func_msgs),
      func_coll_(func_coll)
      {}

    ~IqHandlerRetrieveMessages() {}

    // Shouldn't ever be called.
    virtual bool handleIq(Stanza *stanza) { return false; }

    virtual bool handleIqID(Stanza* stanza, int context)
      {
      if (stanza->error() == StanzaErrorUndefined)
        {
        Tag* retrieve_cmd = stanza->findChild("chat");
        TimeType start = str2DateTime(retrieve_cmd->findAttribute("start"));
        Tag::TagList msgs = retrieve_cmd->children();
        Driver::Messages infos;

        // Pass our single fetched collection - just in case anyone needs it ...
        if (!func_coll_.empty())
          func_coll_(parseCollectionInfo(retrieve_cmd));

        for (Tag::TagList::const_iterator it = msgs.begin(); it != msgs.end(); ++it)
          if ((*it)->name() == "from" || (*it)->name() == "to" || (*it)->name() == "note")
            infos.push_back(parseMessageInfo(*it, start));

        func_msgs_(infos, parseRSM(retrieve_cmd->findChild("set", "xmlns", XMLNS_RESULT_SET_MANAGEMENT)), OK);
        }
      else
        func_msgs_(Driver::Messages(), RSM_Result(), SERVER_ERROR);

      delete this; // Not that nice looking, but simple & effective ;-)
      return true;
      }

  private:
    Driver::ReceiveMessages func_msgs_;
    Driver::ReceiveCollection func_coll_;
  };
//------------------------------------------------------------------------------
class IqHandlerRetrieveModifications: public IqHandler
  {
  public:
    IqHandlerRetrieveModifications(const Driver::ReceiveModifications& func): func_(func) {}

    ~IqHandlerRetrieveModifications() {}

    // Shouldn't ever be called.
    virtual bool handleIq(Stanza *stanza) { return false; }

    virtual bool handleIqID(Stanza* stanza, int context)
      {
      if (stanza->error() == StanzaErrorUndefined)
        {
        Tag* modified_cmd = stanza->findChild("modified");
        Tag::TagList modifs = modified_cmd->children();
        Driver::Modifications infos;

        for (Tag::TagList::const_iterator it = modifs.begin(); it != modifs.end(); ++it)
          if ((*it)->name() == "changed" || (*it)->name() == "removed")
            {
            Modification info;
            info.link = parseCollectionLink(*it);

            if ((*it)->hasAttribute("by"))
              info.by = (*it)->findAttribute("by");

            if ((*it)->name() == "changed")
              info.type = Modification::CHANGED;
            else
              {
              assert((*it)->name() == "removed");
              info.type = Modification::REMOVED;
              }

            infos.push_back(info);
            }

        func_(infos, parseRSM(modified_cmd->findChild("set", "xmlns", XMLNS_RESULT_SET_MANAGEMENT)), OK);
        }
      else
        func_(Driver::Modifications(), RSM_Result(), SERVER_ERROR);

      delete this; // Not that nice looking, but simple & effective ;-)
      return true;
      }

  private:
    Driver::ReceiveModifications func_;
  };
//------------------------------------------------------------------------------
class IqHandlerReceiveStatus: public IqHandler
  {
  public:
    IqHandlerReceiveStatus(const Driver::ReceiveStatus& func): func_(func) {}

    ~IqHandlerReceiveStatus() {}

    // Shouldn't ever be called.
    virtual bool handleIq(Stanza *stanza) { return false; }

    virtual bool handleIqID(Stanza* stanza, int context)
      {
      if (stanza->error() == StanzaErrorUndefined)
        func_(OK);
      else
        func_(SERVER_ERROR);

      delete this; // Not that nice looking, but simple & effective ;-)
      return true;
      }

  private:
    Driver::ReceiveStatus func_;
  };
//------------------------------------------------------------------------------
GlooxDriver::GlooxDriver(gloox::Client* client): client_(client)
  {
  }
//------------------------------------------------------------------------------
void GlooxDriver::initialize(const ReceiveStatus& func)
  {
  }
//------------------------------------------------------------------------------
std::string GlooxDriver::getJID() const
  {
  return client_->jid().full();
  }
//------------------------------------------------------------------------------
void GlooxDriver::retrieveCollections(const Filter& filter, const RSM_Query& rsm, const ReceiveCollections& func)
  {
  // Prepate the appropriate iq request
  Tag* cmd = new Tag("list");
  cmd->addAttribute("xmlns", XMLNS_MESSAGE_ARCHIVING);

  if (!filter.with.empty())
    cmd->addAttribute("with", filter.with);

  if (filter.start != TIME_INVALID)
    cmd->addAttribute("start", dateTime2Str(filter.start));

  if (filter.end != TIME_INVALID)
    cmd->addAttribute("end", dateTime2Str(filter.end));

  addRSM(cmd, rsm);

  sendIqCmd(client_, cmd, new IqHandlerRetrieveCollections(func));
  }
//------------------------------------------------------------------------------
void GlooxDriver::retrieveMessages(const CollectionLink& link, const RSM_Query& rsm, const ReceiveMessages& func_msgs,
                                   const ReceiveCollection& func_coll)
  {
  // Prepate the appropriate iq request
  Tag* cmd = new Tag("retrieve");
  cmd->addAttribute("xmlns", XMLNS_MESSAGE_ARCHIVING);

  cmd->addAttribute("with", link.with);
  cmd->addAttribute("start", dateTime2Str(link.start));

  addRSM(cmd, rsm);

  sendIqCmd(client_, cmd, new IqHandlerRetrieveMessages(func_msgs, func_coll));
  }
//------------------------------------------------------------------------------
void GlooxDriver::retrieveModifications(const Filter& filter, const RSM_Query& rsm, const ReceiveModifications& func)
  {
  // Prepate the appropriate iq request
  Tag* cmd = new Tag("modified");
  cmd->addAttribute("xmlns", XMLNS_MESSAGE_ARCHIVING);

  if (filter.start != TIME_INVALID)
    cmd->addAttribute("start", dateTime2Str(filter.start));

  addRSM(cmd, rsm);

  sendIqCmd(client_, cmd, new IqHandlerRetrieveModifications(func));
  }
//------------------------------------------------------------------------------
void GlooxDriver::updateCollection(const WSW::Core::Collection& collection, const ReceiveStatus& func_status)
  {
  // Prepate the appropriate iq request
  Tag* cmd = new Tag("save");
  cmd->addAttribute("xmlns", XMLNS_MESSAGE_ARCHIVING);

  Tag* chat = new Tag("chat");
  chat->addAttribute("with", collection.info.self.with);
  chat->addAttribute("start", dateTime2Str(collection.info.self.start));

  if (!collection.info.thread.empty())
    chat->addAttribute("thread", collection.info.thread);
  if (!collection.info.subject.empty())
    chat->addAttribute("subject", collection.info.subject);

  if (collection.info.prev.isValid())
    chat->addChild(createLinkTag("previous", collection.info.prev));
  if (collection.info.next.isValid())
    chat->addChild(createLinkTag("next", collection.info.next));

  for (WSW::Core::Collection::Messages::const_iterator it = collection.messages.begin();
       it != collection.messages.end(); ++it)
    {
    Tag* msg = new Tag(it->getDirectionAsString());
    
    if (it->direction == Message::NOTE || it->utc < collection.info.self.start)
      msg->addAttribute("utc", dateTime2Str(it->utc));
    else
      msg->addAttribute("secs", boost::lexical_cast<std::string>((it->utc - collection.info.self.start).total_seconds()));

    if (!it->name.empty())
      msg->addAttribute("name", it->name);

    if (it->direction == Message::NOTE)
      msg->addCData(it->body);
    else
      {
      Tag* body = new Tag("body");
      body->addCData(it->body);
      msg->addChild(body);
      }

    chat->addChild(msg);
    }

  cmd->addChild(chat);

  sendIqCmd(client_, cmd, new IqHandlerReceiveStatus(func_status), StanzaIqSet);
  }
//------------------------------------------------------------------------------
void GlooxDriver::idle()
  {
  client_->recv();
  }
//------------------------------------------------------------------------------
