/*****************************************************************************
 *
 * libwsw, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ejabberd@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef WSW_TASK_CACHED_H
#define WSW_TASK_CACHED_H

#include "api.h"
#include "core.h"
#include "task.h"
//------------------------------------------------------------------------------
// NOTE: do not comment it unless you're sure SOCI fixed integral bindings,
// including boolean type support!
#define USE_SOCI_INTEGRAL_BINDING_HACK
//------------------------------------------------------------------------------
namespace SOCI { class Session; }
//------------------------------------------------------------------------------
namespace WSW
  {
  namespace API_details
    {
//--//--------------------------------------------------------------------------
    struct CachedItem
      {
      static const unsigned long INVALID_ID = (unsigned long)-1;
      unsigned long rsm_index;

      // SQLite3 + SOCI cannot support anything from integral types but int :-(
      // Additionally, SOCI does currently not support binding boolean type at all.
      // Therefore we have to perform temporary caching of these values into int variables.
#ifdef USE_SOCI_INTEGRAL_BINDING_HACK
      int rsm_index_val;
#endif

      CachedItem() { rsm_index = INVALID_ID; }
      };
//--//--------------------------------------------------------------------------
    struct CollectionInfoCached: public CachedItem
      {
      WSW::Core::CollectionInfo coll;
      bool dirty;
      static const unsigned long INVALID_ID = (unsigned long)-1;
      unsigned long prev_coll_id, next_coll_id;

      // SQLite3 + SOCI cannot support anything from integral types but int :-(
      // Additionally, SOCI does currently not support binding boolean type at all.
      // Therefore we have to perform temporary caching of these values into int variables.
    #ifdef USE_SOCI_INTEGRAL_BINDING_HACK
      int dirty_val;
      int crypt_val;
      int prev_coll_id_val, next_coll_id_val;
    #endif

      void clear()
        {
        prev_coll_id = next_coll_id = rsm_index = INVALID_ID;
        }

      CollectionInfoCached() { clear(); }
      CollectionInfoCached(const WSW::Core::CollectionInfo& info): coll(info) { clear(); }
      };
//--//--------------------------------------------------------------------------
  struct MessageCached: public CachedItem
    {
    WSW::Core::Message msg;

    // SQLite3 + SOCI cannot support anything from integral types but int :-(
    // Additionally, SOCI does currently not support binding boolean type at all.
    // Therefore we have to perform temporary caching of these values into int variables.
#ifdef USE_SOCI_INTEGRAL_BINDING_HACK
    int direction_val;
#endif

    void clear()
      {
      rsm_index = INVALID_ID;
      }

    MessageCached() { clear(); }
    MessageCached(const WSW::Core::Message& info): msg(info) { clear(); }
    };
//--//--------------------------------------------------------------------------
    class TaskCached: public virtual Task
      {
      public:
        TaskCached(WSW::Drivers::XMPP::Driver& driver, const WSW::API::Options& opts,
                   const WSW::API::ReceiveStatus& func_status);

        // This has to be separate to maintain common constructor interface between Task and TaskCached
        void setSession(SOCI::Session* session) { session_ = session; }

        virtual std::string getLockingJID() = 0;
        virtual void performGetItemsNumberCacheQuery() = 0;

        virtual void init();
        virtual void cancel(StatusAPI res);
        virtual void finished();

        virtual void postReceiveModifications() {}

        bool startTransaction(const ResumeFunction& func = ResumeFunction());
        void transactionFinished();
        bool tryLockJID(const std::string& jid, const ResumeFunction& func);
        void unlockJID(const std::string& jid);
        bool tryLockForModificationCmd(const ResumeFunction& func);
        void unlockForModificationCmd();
        void getReplicationInfo();
        void setupReceiveModifications();
        void updateIndicesAfter(const WSW::Core::CollectionLink& in_link, long amount);
        unsigned long getCollectionIdByLink(SOCI::eIndicator& ind, const WSW::Core::CollectionLink& in_link);
        void createCollectionByLink(const WSW::Core::CollectionLink& in_link, bool fl_calculate_index = true);
        unsigned long unResolveLink(const WSW::Core::CollectionLink& link);
        void unResolveLinks(CollectionInfoCached& info);
        void updateCollectionInCache(const CollectionInfoCached& in_info, unsigned long coll_id);
        void updateMessageInCache(const MessageCached& msg, unsigned long coll_id);
        bool cacheIsCompleteForLink(const std::string& with, const WSW::DateTime::TimeType& start);
        void addNewCollectionByLink(const std::string& with, const WSW::DateTime::TimeType& start);
        void processChangedCollectionLink(const std::string& with, const WSW::DateTime::TimeType& start);
        void removeCollectionByLink(const std::string& with, const WSW::DateTime::TimeType& start);
        void forceReplicationOnNextRequest();
        void resetCollectionsIndices(const WSW::Core::CollectionLink& link_in);
        void receiveModifications(const WSW::Drivers::XMPP::Driver::Modifications& modifs,
                                  const WSW::Core::RSM_Result& rsm, StatusDriver status);

      protected:
        SOCI::Session* session_;
        
        bool transaction_started_;

        WSW::DateTime::TimeType last_update_, replication_start_;
        bool fl_updated_, fl_replication_needed_, fl_replication_in_progress_;
        SOCI::eIndicator ind_last_update_;
        unsigned long modified_items_, modified_count_;
        WSW::DateTime::TimeType modified_start_;
        std::string modified_last_;

        SOCI::eIndicator ind_items_number_;

        std::string jid_;
      };
    }
  }
//------------------------------------------------------------------------------
namespace SOCI
  {
  template<> struct TypeConversion<WSW::API_details::CollectionInfoCached>
    {
    typedef Values base_type;

    static WSW::API_details::CollectionInfoCached from(Values& v)
      {
      WSW::API_details::CollectionInfoCached info;

      info.coll.self.with = v.get<std::string>("with_jid");
      info.coll.self.start = v.get<WSW::DateTime::TimeType>("utc");
      info.coll.subject = v.get<std::string>("subject", "");
      info.coll.thread = v.get<std::string>("thread", "");
      info.coll.extra = v.get<std::string>("extra", "");

#ifdef USE_SOCI_INTEGRAL_BINDING_HACK
      info.coll.crypt = v.get<int>("crypt", 0) != 0;
      info.dirty = v.get<int>("dirty", 1) != 0;
      info.rsm_index = v.get<int>("rsm_index", 0);
      info.prev_coll_id = v.get<int>("prev_coll_id", WSW::API_details::CollectionInfoCached::INVALID_ID);
      info.next_coll_id = v.get<int>("next_coll_id", WSW::API_details::CollectionInfoCached::INVALID_ID);
#else
      info.coll.crypt = v.get<bool>("crypt", false);
      info.coll.dirty = v.get<bool>("dirty", true);
      info.rsm_index = v.get<unsigned long>("rsm_index", 0);
      info.prev_coll_id = v.get<unsigned long>("prev_coll_id", WSW::API_details::CollectionInfoCached::INVALID_ID);
      info.next_coll_id = v.get<unsigned long>("next_coll_id", WSW::API_details::CollectionInfoCached::INVALID_ID);
#endif

      return info;
      }

    static Values to(WSW::API_details::CollectionInfoCached& info)
      {
      Values v;

      v.set("with_jid", info.coll.self.with);
      v.set("utc", info.coll.self.start);
      v.set("subject", info.coll.subject);
      v.set("thread", info.coll.thread);
      v.set("extra", info.coll.extra);

#ifdef USE_SOCI_INTEGRAL_BINDING_HACK
      info.crypt_val = (int)info.coll.crypt; v.set("crypt", info.crypt_val);
      info.dirty_val = (int)info.dirty; v.set("dirty", info.dirty_val);
      info.rsm_index_val = (int)info.rsm_index; v.set("rsm_index", info.rsm_index_val);
      info.prev_coll_id_val = (int)info.prev_coll_id;
      v.set("prev_coll_id", info.prev_coll_id_val, info.prev_coll_id == WSW::API_details::CollectionInfoCached::INVALID_ID? eNull: eOK);
      info.next_coll_id_val = (int)info.next_coll_id;
      v.set("next_coll_id", info.next_coll_id_val, info.next_coll_id == WSW::API_details::CollectionInfoCached::INVALID_ID? eNull: eOK);
#else
      v.set("crypt", info.coll.crypt);
      v.set("dirty", info.dirty);
      v.set("rsm_index", info.rsm_index);
      v.set("prev_coll_id", info.prev_coll_id, info.prev_coll_id == WSW::API_details::CollectionInfoCached::INVALID_ID? eNull: eOK);
      v.set("next_coll_id", info.next_coll_id, info.next_coll_id == WSW::API_details::CollectionInfoCached::INVALID_ID? eNull: eOK);
#endif

      return v;
      }
    };
  }
//------------------------------------------------------------------------------
namespace SOCI
  {
  template<> struct TypeConversion<WSW::API_details::MessageCached>
    {
    typedef Values base_type;

    static WSW::API_details::MessageCached from(Values& v)
      {
      WSW::API_details::MessageCached info;

      info.msg.utc = v.get<WSW::DateTime::TimeType>("utc");
      info.msg.body = v.get<std::string>("body", "");
      info.msg.name = v.get<std::string>("name", "");

#ifdef USE_SOCI_INTEGRAL_BINDING_HACK
      info.msg.direction = WSW::Core::Message::Direction(v.get<int>("direction"));
      info.rsm_index = v.get<int>("rsm_index", 0);
#else
      info.msg.direction = v.get<Message::Direction>("direction");
      info.rsm_index = v.get<unsigned long>("rsm_index", 0);
#endif

      return info;
      }

    static Values to(WSW::API_details::MessageCached& info)
      {
      Values v;

      v.set("utc", info.msg.utc);
      v.set("body", info.msg.body);
      v.set("name", info.msg.name);

#ifdef USE_SOCI_INTEGRAL_BINDING_HACK
      info.direction_val = (int)info.msg.direction; v.set("direction", info.direction_val);
      info.rsm_index_val = (int)info.rsm_index; v.set("rsm_index", info.rsm_index_val);
#else
      v.set("direction", info.msg.direction);
      v.set("rsm_index", info.rsm_index);
#endif

      return v;
      }
    };
  }
//------------------------------------------------------------------------------

#endif // WSW_TASK_INDEXED_CACHED_H