/*****************************************************************************
 *
 * libwsw, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ejabberd@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef WSW_TASK_INDEXED_H
#define WSW_TASK_INDEXED_H

#include "api.h"
#include "task.h"
//------------------------------------------------------------------------------
namespace WSW
  {
  namespace API_details
    {
    class TaskIndexed: public virtual Task
      {
      public:
        TaskIndexed(WSW::Drivers::XMPP::Driver& driver, const WSW::API::Options& opts,
                    const WSW::API::ReceiveStatus& func_status, long first_index, long last_index);
        virtual ~TaskIndexed() {}

        virtual void setupReceiveItemsQuery() = 0;
        virtual void setupGetItemsNumberQuery() = 0;

        // Children need to reimplement it only if some additional actions need to be performed, such as
        // updating database or whatever.
        virtual void postReceiveItemsNumber() {}

        // These default implementations should work OK for all direct tasks
        virtual void init();
        virtual void fetchItems();
        virtual void receiveItemsQueryFinished();

        // As we do not have templatized virtual functions, we'll have to list all possible types here.
        // It's not very clean way, but fortunately, there's not that many types ;-)
        // (we could also do it via common base class, but it would be much uglier, I think)
        virtual void receiveItemsHandleItem(const WSW::Core::CollectionInfo& /*info*/, unsigned long /*index*/) {}
        virtual void receiveItemsHandleItem(const WSW::Core::Message& /*msg*/, unsigned long /*index*/) {}

        virtual bool verifyRSM(const WSW::Core::RSM_Result& rsm, unsigned long received_size);

        bool calculateIndices();
        WSW::Core::RSM_Query setupRSM_ForItemsQuery(unsigned long limit);
        void postReceiveItems(const WSW::Core::RSM_Result& rsm, unsigned long received_size);
        void setupGetItemsNumber();

        template <class Items>
        void receiveItems(const Items& items, const WSW::Core::RSM_Result& rsm, StatusDriver status);

        template <class Items>
        void receiveItemsNumber(const Items& items, const WSW::Core::RSM_Result& rsm, StatusDriver status);

      protected:
        bool fl_reversed_, fl_items_len_defined_;
        long first_index_, last_index_, first_, last_, first_item_index_;
        unsigned long items_len_, items_received_;
        std::string item_last_;
      };
    }
  }
//------------------------------------------------------------------------------
#endif // WSW_TASK_INDEXED_H
